#include "nghds.h"

/*
 * Function to form the neighbourhoods for a pedigree with data matrix formed
 * by the input_data routine. The mother and father submatrices are formed
 * here and a sex check caried out. Existence of the nghd array is assumed
 * and these structures are all initialised to be null. The null node also
 * has to be initialised so that we have somewhere to send the unknown
 * parents of founders. 
 * Routine by Charlie Geyer (1991) based on earlier routine of Nuala Sheehan.
 *
 * component indicator of individuals added; Sept 1993; EAT. 
 */

static int **mother;			/* was mother[MAX_SIZE][3] */
static int **father;			/* was father[MAX_SIZE][3] */

static int father_mat(int **, int **), mother_mat(int **, int **);
static void bounds(int, int **, int *), nullify(Ind *); 
static void parent_sort(int **), make_nghd(int **);
static void node(int, int **, int *), smallnode(int), verysmallnode(int);

void nghds(int **data_matrix)
{
    int i, j, bad_dads, bad_mums;

    mother = (int **) malloc((pedsize + 1) * sizeof(int *));
    alloc_chk(WHERE, mother);
    mother[0] = (int *) malloc((pedsize + 1) * 3 * sizeof(int));
    alloc_chk(WHERE, mother[0]);
    for (i = 1; i <= pedsize; i++)
	mother[i] = mother[i - 1] + 3;

    father = (int **) malloc((pedsize + 1) * sizeof(int *));
    alloc_chk(WHERE, father);
    father[0] = (int *) malloc((pedsize + 1) * 3 * sizeof(int));
    alloc_chk(WHERE, father[0]);
    for (i = 1; i <= pedsize; i++)
	father[i] = father[i - 1] + 3;

    bad_dads = father_mat(data_matrix, father);
    parent_sort(father);

    bad_mums = mother_mat(data_matrix, mother);
    parent_sort(mother);

    if (bad_dads > 0 || bad_mums > 0)
        if (bad_dads + bad_mums == 1)
	   give_up_where(WHERE, ERN_sx1, bad_dads + bad_mums); 
        else
	   give_up_where(WHERE, ERN_sex, bad_dads + bad_mums); 

    /* Initialise all neighbourhoods to null. */

    for (i = 0; i <= pedsize; i++)
	nullify(&nghd[i]);
    nullify(&nullnode);

    make_nghd(father);
    make_nghd(mother);

    /*
     * Now we make the neighbourhoods for those who are not parents i. e. 
     * for the individuals at the bottom of the pedigree. Single-parent families
     * are ignored at present. 
     */

    for (j = 0; j <= namestring; j++)
    {
	int row_index = posit[j][0];

	if (posit[j][1] == 0 && row_index != INVALID_INDEX)
	    if (data[row_index][2] == UNKNOWN && data[row_index][3] == UNKNOWN)
		verysmallnode(j);
	    else
	    if (data[row_index][2] == UNKNOWN || data[row_index][3] == UNKNOWN)
	        give_up_where(WHERE, ERN_1pt, j);
	    else
		smallnode(j);
    }
    free((void *) mother[0]);
    free((void *) mother);
    free((void *) father[0]);
    free((void *) father);

    /* Now check that everybody has a nghd so posit[i][1] is pointless */

    for (i = 0; i < pedsize; i++)
    {
	int name = data[i][1];
  	if (posit[name][1] != 1) 
	    give_up_where(WHERE, ERN_ngh, name);
    }
}

/*
 * Function to make the neighbourhoods for all the individuals
 * in parent (i.e.the sorted mother/father matrix). 
 * It calls on the functions bounds() and node() and assumes 
 * the existence of the postion array posit[namestring][2], 
 * of which we only use the first namestring rows. 
 */

static void make_nghd(int **parent)
{
    int k, limits[2], name, parent_index;

    /*
     * Run down through all the parents coded as 0 i. e. founders' parents. 
     */
    for (k = 0; k < pedsize && parent[k][1] == UNKNOWN; k++)
	;
    parent_index = k;

    while (parent_index < pedsize)
    {
	/* name of parent */
	name = parent[parent_index][1];

	bounds(parent_index, parent, limits);

	/*
	 * Now have the two delimiting indices of the block of children
	 * associated with the individual indexed by parent_index. 
	 */

	if (posit[name][1] == 0)
	    /* Have no neighbourhood yet */
	    node(name, parent, limits);
	else
            give_warn(ERN_mng, "make_nghd", name);

	if (limits[1] < pedsize)
	    parent_index = limits[1] + 1;
    }
}

/*
 * Function to find the upper and lower indices of the block of children
 * associated with a given individual whose index in a, "number",
 * is already known.
 * The array a is the sorted mother or father submatrix formed by the
 * functions father_mat() and mother_mat() respectively.
 * These indices are put into a length-2 array, "limits". 
 * The individual of interest is in the second column of the array, a,
 * and the spice in the third column are then sorted within the indices 
 * returned.
 */

static void bounds(int number, int **a, int limits[2])
{
    int i, j, k, temp1, temp2, temp3;

    for (i = number; i + 1 < pedsize && a[i + 1][1] == a[number][1]; i++)
	;
    limits[1] = i;

    for (i = number; i - 1 >= 0 && a[i - 1][1] == a[number][1]; i--)
	;
    limits[0] = i;

    /* sort spice within bounds found above. */

    for (j = limits[0]; j < limits[1]; j++)
	for (k = j + 1; k < limits[1] + 1; ++k)
	    if (a[j][2] > a[k][2])
	    {
		temp1 = a[j][0];
		temp2 = a[j][1];
		temp3 = a[j][2];
		a[j][0] = a[k][0];
		a[j][1] = a[k][1];
		a[j][2] = a[k][2];
		a[k][0] = temp1;
		a[k][1] = temp2;
		a[k][2] = temp3;
	    }
}

/*
 * Function to find the neighbourhood of an individual in the pedigree who
 * is, himself, a parent. 
 * The function takes three arguments: the name of the individual, the 
 * sorted father or mother submatrix, "parent", (depending on whether
 * the individual is male or female), and the 2-dimensional array "limits"
 * returned by the "bounds" function giving the upper and lower indices
 * of the block of children in "parent" associated with the individual
 * in question.  
 * The data matrix data[pedsize + 1][NcolData], position vector 
 * posit[namestring][2], sorted father or mother submatrix 
 * parent[pedsize + 1][3], * and the array of structures of type ind, 
 * nghd[pedsize + 1], are all available.  We also have the index of the
 * individual in parent (i.e.sorted father/ mother matrix), denoted by 
 * parent_index.  
 * It is assumed that the null node has been set up as a global variable
 * with its "self" component initialised to zero to cope with calls to
 * parents of founders. 
 */

static void node(int name, int **parent, int limits[2])
{
    Indlist *child_ptr;
    Marlist *marriage_ptr;
    int row_index, i, j;

    row_index = posit[name][0];

    /* check for a valid index */
    if (row_index != INVALID_INDEX)
    {

	nghd[row_index].self = name;
	nghd[row_index].index = row_index;
	nghd[row_index].component = -1;
	if (data[row_index][2] == UNKNOWN)
	{
	    /* no father */
	    nghd[row_index].pa = &nullnode;
	} else
	{
	    if (posit[data[row_index][2]][0] != INVALID_INDEX)
	    {
		nghd[row_index].pa = &nghd[posit[data[row_index][2]][0]];
	    } else
	        give_up_where(WHERE, ERN_xpa, "node", name);
	}

	if (data[row_index][3] == UNKNOWN)
	{
	    /* no mother */
	    nghd[row_index].ma = &nullnode;
	} else
	{
	    if (posit[data[row_index][3]][0] != INVALID_INDEX)
	    {
		nghd[row_index].ma = &nghd[posit[data[row_index][3]][0]];
	    } else
	        give_up_where(WHERE, ERN_xma, "node", name);
	}

	/*
	 * Now have limits[0] and limits[1] s. t. 
	 * parent[limits[0][0] is the row index in the data matrix for the
	 * first child, and parent[limits[0]][2] is the name of the other parent. 
	 * The matrix "parent" will be the sorted mother/father submatrix. 
	 * We now set up the linked list of marriages and the linked list of
	 * children within each marriage allocating space for the appropriate
	 * structures when required. 
	 */

	i = limits[0];
	nghd[row_index].marriages = (Marlist *) malloc(sizeof(Marlist));
	alloc_chk(WHERE, nghd[row_index].marriages);
	nghd[row_index].marriages->current = (Mar *) malloc(sizeof(Mar));
	alloc_chk(WHERE, nghd[row_index].marriages->current);
	nghd[row_index].marriages->current->spouse = &nghd[posit[parent[i][2]][0]];
	marriage_ptr = nghd[row_index].marriages;

	while (i < limits[1] + 1)
	{
	    marriage_ptr->current->kids = (Indlist *) malloc(sizeof(Indlist));
	    alloc_chk(WHERE, marriage_ptr->current->kids);
	    marriage_ptr->current->kids->current = &nghd[parent[i][0]];
	    child_ptr = marriage_ptr->current->kids;

	    for (j = i + 1; j < limits[1] + 1 && parent[j][2] == parent[i][2]; ++j)
	    {
		child_ptr->next = (Indlist *) malloc(sizeof(Indlist));
		alloc_chk(WHERE, child_ptr->next);
		child_ptr = child_ptr->next;
		child_ptr->current = &nghd[parent[j][0]];
	    }
	    child_ptr->next = NULL;

	    i = j;			/* Increment for the while loop */

	    if (i != limits[1] + 1)
	    {
		marriage_ptr->next = (Marlist *) malloc(sizeof(Marlist));
		alloc_chk(WHERE, marriage_ptr->next);
		marriage_ptr = marriage_ptr->next;
		marriage_ptr->current = (Mar *) malloc(sizeof(Mar));
		alloc_chk(WHERE, marriage_ptr->current);
		marriage_ptr->current->spouse = &nghd[posit[parent[i][2]][0]];
	    }
	}				/* end of while */

	marriage_ptr->next = NULL;

	/* Indicates that neighbourhood is made */
	posit[name][1] = 1;
    } else
	/* Have an invalid row index - individual does not exist */
	give_up_where(WHERE, ERN_nop, "node", row_index); 
}

/*
 * Function to find the "restricted" neighbourhood for an individual who 
 * does not feature as a father or a mother. Since no children are involved, 
 * the spice (if such exist) are of no genetic interest, and so we declare 
 * the marriage to be null and void. The position array posit[namestring][2] 
 * is available and these last individuals are found to be those with a zero
 * entry in the second column of this array. 
 * The input for the routine is the name of the individual in question.
 * It is assumed that a valid pedigree name has been supplied to the function,
 * and that any screening has occured prior to the functin call.
 */

static void smallnode(int name)
{
    int row_index;

    row_index = posit[name][0];

    nghd[row_index].self = name;
    nghd[row_index].index = row_index;
    nghd[row_index].component = -1;
    nghd[row_index].pa = &nghd[posit[data[row_index][2]][0]];
    nghd[row_index].ma = &nghd[posit[data[row_index][3]][0]];
    nghd[row_index].marriages = NULL;

    /* Indicate that a neighbourhood has been made */
    posit[name][1] = 1;
}

/*
 * Function to find the "very restricted" neighbourhood for an individual who
 * does not feature as a father or a mother and has no spouses or children. 
 *
 * Since no children are involved, the spice (if such exist) are of no genetic 
 * interest, and so we declare the marriage to be null and void. The position 
 * array posit[namestring][2] is available and these last individuals are found 
 * to be those with a zero entry in the second column of this array. 
 * The input for the routine is the name of the individual in question.
 * It is assumed that a valid pedigree name has been supplied to the function, 
 * and that any screening has occured prior to the function call. 
 */

static void verysmallnode(int name)
{
    int row_index;

    row_index = posit[name][0];

    nghd[row_index].self = name;
    nghd[row_index].index = row_index;
    nghd[row_index].component = -1;
    nghd[row_index].pa = &nullnode;
    nghd[row_index].ma = &nullnode;
    nghd[row_index].marriages = NULL;

    /* Indicate that a neighbourhood has been made */
    posit[name][1] = 1;
}

/*
 * Function to bubble-sort the second column of a 3-column matrix in
 * ascending order(i.e.smallest in first row), and to apply the same
 * permutation to the first and third columns. This is used to sort the
 * mother submatrix on mothers and the father submatrix on fathers. 
 */

static void parent_sort(int **a)
{
    int i, j, temp1, temp2, temp3;

    for (i = 0; i < pedsize - 1; i++)
	for (j = i + 1; j < pedsize; j++)
	    if (a[i][1] > a[j][1])
	    {
		temp1 = a[i][0];
		temp2 = a[i][1];
		temp3 = a[i][2];
		a[i][0] = a[j][0];
		a[i][1] = a[j][1];
		a[i][2] = a[j][2];
		a[j][0] = temp1;
		a[j][1] = temp2;
		a[j][2] = temp3;
	    }
}

/*
 * Function to make the nullnode for the neighbourhood routine. 
 * Pointers to pa, ma and marriages are set to NULL. 
 * It is assumed that a structure of type Ind has already been set up
 * nullnode and defined to be a global variable.
 */

static void nullify(Ind * nul)
{
    nul->self = UNKNOWN;
    nul->index = INVALID_INDEX;
    nul->pa = NULL;
    nul->ma = NULL;
    nul->marriages = NULL;
}

/*
 * Function to create the mother submatrix consisting of the index column,
 * the column of mother labels and the column of father labels.
 * A sex check is included at this point. 
 * The function returns the number of mothers who have been given the 
 * wrong sex.  (One case of mistaken sex could feature for as many times 
 * as she has children.) 
 * This matrix will then be sorted by mothers, the same permutation being
 * applied to the other two columns, which will enable us to pick off 
 * the list of children associated with each mother.  (These are given by
 * the indices in the first column.)  The third column, the fathers, will
 * later be sorted by the bounds function for each mother.
 * The program will not die at this point if there is a mistake with gender.
 * It is of interest to find how many such mistakes there are.
 * The nghds function will die if this function returns a * positive number. 
 */

static int mother_mat(int **matrix, int **mot)
{
    int i;
    int n = 0;

    for (i = 0; i < pedsize; i++)
    {
	mot[i][0] = matrix[i][0];
	mot[i][1] = matrix[i][3];
	mot[i][2] = matrix[i][2];

	/* Checking gender is female. */

	if (mot[i][1] != UNKNOWN && matrix[posit[mot[i][1]][0]][4] != FEMALE)
	{
            give_warn_where(WHERE, ERN_sxm, "mother_mat", mot[i][1]);
	    ++n;
	}
    }
    return n;
}

/*
 * Function to create the father submatrix consisting of the index column,
 * the column of father labels and the column of mother labels. 
 * A sex check is included at this point. 
 * The function returns the number of fathers who have been given the 
 * wrong sex.  (One case of mistaken sex could feature for as many times
 * as he has children.) 
 * This matrix will then be sorted by fathers, the same permutatiion being
 * applied to the other two columns, which will enable us to pick off
 * the list of children associated with each father.  (These are given by
 * the indices in the first column.)  The third column, the mothers, will
 * later be sorted by the bounds function for each father.  
 * The program will not die at this point if there is a mistake with gender.
 * It is of interest to find how many such mistakes there are.  
 * The nghds function will die if this function returns a * positive number. 
 */

static int father_mat(int **matrix, int **fat)
{
    int i;
    int n = 0;

    for (i = 0; i < pedsize; i++)
    {

	fat[i][0] = matrix[i][0];
	fat[i][1] = matrix[i][2];
	fat[i][2] = matrix[i][3];

	/* Checking gender is male. */
	if (fat[i][1] != UNKNOWN && matrix[posit[fat[i][1]][0]][4] != MALE)
	{
            give_warn_where(WHERE, ERN_sxf, "father_mat", fat[i][1]);
	    ++n;
	}
    }
    return n;
}

